/*
 * $Id: ship.d,v 1.2 2005/07/03 07:05:22 kenta Exp $
 *
 * Copyright 2005 Kenta Cho. Some rights reserved.
 */
module abagames.gr.ship;

private import std.math;
private import opengl;
private import abagames.util.vector;
private import abagames.util.rand;
private import abagames.util.math;
private import abagames.util.sdl.pad;
private import abagames.util.sdl.twinstick;
private import abagames.util.sdl.recordableinput;
private import abagames.util.sdl.shape;
private import abagames.gr.field;
private import abagames.gr.gamemanager;
private import abagames.gr.screen;
private import abagames.gr.particle;
private import abagames.gr.letter;
private import abagames.gr.shot;
private import abagames.gr.enemy;
private import abagames.gr.stagemanager;
private import abagames.gr.soundmanager;
private import abagames.gr.prefmanager;
private import abagames.gr.shape;

/**
 * Player's ship.
 */
public class Ship {
 private:
  static const int RESTART_CNT = 300;
  static const int INVINCIBLE_CNT = 228;
  static const float HIT_WIDTH = 0.02f;
  static const int FIRE_INTERVAL = 2;
  static const int FIRE_INTERVAL_MAX = 4;
  static const int FIRE_LANCE_INTERVAL = 15;
  static const float SCROLL_SPEED_BASE = 0.01f;
  static const float SCROLL_SPEED_MAX = 0.1f;
  static const float SCROLL_START_Y = 2.5f;
  static const float SPEED_BASE = 0.15f;
  static const float TURN_RATIO_BASE = 0.2f;
  static const float SLOW_TURN_RATIO = 0;
  static const float TURN_CHANGE_RATIO = 0.5f;
  static Rand rand;
  RecordablePad pad;
  RecordableTwinStick twinStick;
  Field field;
  Screen screen;
  ShotPool shots;
  SparkPool sparks;
  SmokePool smokes;
  FragmentPool fragments;
  WakePool wakes;
  EnemyPool enemies;
  StageManager stageManager;
  InGameState gameState;
  Vector _pos;
  Vector firePos;
  float deg;
  float speed;
  float turnRatio;
  BaseShape _shape;
  BaseShape bridgeShape;
  int fireCnt;
  int fireSprCnt;
  float fireInterval;
  float fireSprDeg;
  int fireLanceCnt;
  float fireDeg;
  bool aPressed, bPressed;
  int cnt;
  float scrollSpeed, _scrollSpeedBase;
  bool onBlock;
  Vector _vel;
  Vector refVel;
  int shieldCnt;
  ShieldShape shieldShape;
  bool _replayMode;
  float turnSpeed;
  bool reverseFire;
  int gameMode;
  PadState padInput;
  TwinStickState stickInput;
  float vx, vy;

  invariant {
    assert(_pos.x < 15 && _pos.x > -15);
    assert(_pos.y < 20 && _pos.y > -20);
    assert(firePos.x < 15 && firePos.x > -15);
    assert(firePos.y < 20 && firePos.y > -20);
    assert(deg <>= 0);
    assert(speed >= 0 && speed <= SPEED_BASE);
    assert(turnRatio >= 0 && turnRatio <= TURN_RATIO_BASE);
    assert(fireInterval >= FIRE_INTERVAL);
    assert(fireSprDeg <>= 0);
    assert(cnt >= -RESTART_CNT);
    assert(scrollSpeed > 0);
    assert(_scrollSpeedBase > 0);
    assert(_vel.x < 2 && _vel.x > -2);
    assert(_vel.y < 2 && _vel.y > -2);
    assert(refVel.x < 1 && refVel.x > -1);
    assert(refVel.y < 1 && refVel.y > -1);
    assert(shieldCnt >= 0);
  }

  public this(Pad pad, TwinStick twinStick, Field field, Screen screen,
              SparkPool sparks, SmokePool smokes, FragmentPool fragments, WakePool wakes) {
    rand = new Rand;
    this.pad = cast(RecordablePad) pad;
    this.twinStick = cast(RecordableTwinStick) twinStick;
    this.field = field;
    this.screen = screen;
    this.sparks = sparks;
    this.smokes = smokes;
    this.fragments = fragments;
    this.wakes = wakes;
    _pos = new Vector;
    firePos = new Vector;
    _vel = new Vector;
    refVel = new Vector;
    _shape = new BaseShape(0.7f, 0.6f, 0.6f, BaseShape.ShapeType.SHIP_ROUNDTAIL, 0.5f, 0.7f, 0.5f);
    bridgeShape = new BaseShape(0.3f, 0.6f, 0.6f, BaseShape.ShapeType.BRIDGE, 0.3f, 0.7f, 0.3f);
    deg = 0;
    speed = 0;
    turnRatio = 0;
    fireInterval = FIRE_INTERVAL;
    fireSprDeg = 0;
    cnt = 0;
    scrollSpeed = _scrollSpeedBase = SCROLL_SPEED_BASE;
    shieldCnt = 0;
    shieldShape = new ShieldShape;
  }

  public void setRandSeed(long seed) {
    rand.setSeed(seed);
  }

  public void close() {
    _shape.close();
    bridgeShape.close();
    shieldShape.close();
  }

  public void setShots(ShotPool shots) {
    this.shots = shots;
  }

  public void setEnemies(EnemyPool enemies) {
    this.enemies = enemies;
  }

  public void setStageManager(StageManager stageManager) {
    this.stageManager = stageManager;
  }

  public void setGameState(InGameState gameState) {
    this.gameState = gameState;
  }

  public void start(int gameMode) {
    this.gameMode = gameMode;
    _pos.x = 0;
    _pos.y = -field.size.y * 0.8f;
    firePos.x = firePos.y = 0;
    _vel.x = _vel.y = 0;
    deg = 0;
    speed = SPEED_BASE;
    turnRatio = TURN_RATIO_BASE;
    cnt = -INVINCIBLE_CNT;
    _scrollSpeedBase = SCROLL_SPEED_BASE;
    aPressed = bPressed = true;
    restart();
  }

  public void restart() {
    switch (gameMode) {
    case InGameState.GameMode.NORMAL:
      fireCnt = 99999;
      fireInterval = 99999;
      break;
    case InGameState.GameMode.TWIN_STICK:
      fireCnt = 0;
      fireInterval = FIRE_INTERVAL;
      break;
    }
    fireSprCnt = 0;
    fireSprDeg = 0.5f;
    fireLanceCnt = 0;
    scrollSpeed = _scrollSpeedBase;
    if (field.getBlock(_pos) >= 0)
      onBlock = true;
    else
      onBlock = false;
    refVel.x = refVel.y = 0;
    shieldCnt = 20 * 60;
  }

  public void move() {
    float px = _pos.x, py = _pos.y;
    field.scroll(scrollSpeed);
    cnt++;
    vx = vy = 0;
    switch (gameMode) {
    case InGameState.GameMode.NORMAL:
      moveNormal();
      break;
    case InGameState.GameMode.TWIN_STICK:
      moveTwinStick();
      break;
    }
    if (gameState.isGameOver) {
      clearBullets();
      if (cnt < -INVINCIBLE_CNT)
        cnt = -RESTART_CNT;
    } else if (cnt < -INVINCIBLE_CNT) {
      clearBullets();
    }
    vx *= speed;
    vy *= speed;
    vx += refVel.x;
    vy += refVel.y;
    refVel *= 0.9f;
    if (field.checkInField(_pos.x, _pos.y - field.lastScrollY))
      _pos.y -= field.lastScrollY;
    if ((onBlock || field.getBlock(_pos.x + vx, _pos.y) < 0) &&
        field.checkInField(_pos.x + vx, _pos.y)) {
      _pos.x += vx;
      _vel.x = vx;
    } else {
      _vel.x = 0;
      refVel.x = 0;
    }
    bool srf = false;
    if ((onBlock || field.getBlock(px, _pos.y + vy) < 0) &&
        field.checkInField(_pos.x, _pos.y + vy)) {
      _pos.y += vy;
      _vel.y = vy;
    } else {
      _vel.y = 0;
      refVel.y = 0;
    }
    if (field.getBlock(_pos.x, _pos.y) >= 0) {
      if (!onBlock)
        if (cnt <= 0)
          onBlock = true;
        else {
          if (field.checkInField(_pos.x, _pos.y - field.lastScrollY)) {
            _pos.x = px;
            _pos.y = py;
          } else {
            destroyed();
          }
        }
    } else {
      onBlock = false;
    }
    if (_pos.y >= SCROLL_START_Y) {
      scrollSpeed += (SCROLL_SPEED_MAX - scrollSpeed) * 0.1f;
    } else {
      scrollSpeed += (_scrollSpeedBase - scrollSpeed) * 0.1f;
    }
    switch (gameMode) {
    case InGameState.GameMode.NORMAL:
      fireNormal();
      break;
    case InGameState.GameMode.TWIN_STICK:
      fireTwinStick();
      break;
    }
    _scrollSpeedBase += (SCROLL_SPEED_MAX - _scrollSpeedBase) * 0.00001f;
    if (cnt % 3 == 0 && cnt >= -INVINCIBLE_CNT) {
      float sp;
      if (vx != 0 || vy != 0)
        sp = 0.4f;
      else
        sp = 0.2f;
      sp *= 1 + rand.nextSignedFloat(0.33f);
      sp *= SPEED_BASE;
      _shape.addWake(wakes, _pos, deg, sp);
    }
    Enemy he = enemies.checkHitShip(pos.x, pos.y);
    if (he) {
      float rd;
      if (pos.dist(he.pos) < 0.1f)
        rd = 0;
      else
        rd = atan2(_pos.x - he.pos.x, _pos.y - he.pos.y);
      assert(rd <>= 0);
      float sz = he.size;
      refVel.x = sin(rd) * sz * 0.1f;
      refVel.y = cos(rd) * sz * 0.1f;
      float rs = refVel.vctSize;
      if (rs > 1) {
        refVel.x /= rs;
        refVel.y /= rs;
      }
    }
    if (shieldCnt > 0)
      shieldCnt--;
    if (cnt >= -INVINCIBLE_CNT && _pos.x > field.size.x / 3 && _pos.y < -field.size.y / 4 * 3)
      gameState.shrinkScoreReel();
  }

  private void moveNormal() {
    if (!_replayMode) {
      padInput = pad.getState();
    } else {
      try {
        padInput = pad.replay();
      } catch (NoRecordDataException e) {
        gameState.isGameOver = true;
        padInput = pad.getNullState();
      }
    }
    if (gameState.isGameOver || cnt < -INVINCIBLE_CNT)
      padInput.clear();
    if (padInput.dir & PadState.Dir.UP)
      vy = 1;
    if (padInput.dir & PadState.Dir.DOWN)
      vy = -1;
    if (padInput.dir & PadState.Dir.RIGHT)
      vx = 1;
    if (padInput.dir & PadState.Dir.LEFT)
      vx = -1;
    if (vx != 0 && vy != 0) {
      vx *= 0.7f;
      vy *= 0.7f;
    }
    if (vx != 0 || vy != 0) {
      float ad = atan2(vx, vy);
      assert(ad <>= 0);
      Math.normalizeDeg(ad);
      ad -= deg;
      Math.normalizeDeg(ad);
      deg += ad * turnRatio * turnSpeed;
      Math.normalizeDeg(deg);
    }
  }

  private void moveTwinStick() {
    if (!_replayMode) {
      stickInput = twinStick.getState();
    } else {
      try {
        stickInput = twinStick.replay();
      } catch (NoRecordDataException e) {
        gameState.isGameOver = true;
        stickInput = twinStick.getNullState();
      }
    }
    if (gameState.isGameOver || cnt < -INVINCIBLE_CNT)
      stickInput.clear();
    vx = stickInput.left.x;
    vy = stickInput.left.y;
    if (vx != 0 || vy != 0) {
      float ad = atan2(vx, vy);
      assert(ad <>= 0);
      Math.normalizeDeg(ad);
      ad -= deg;
      Math.normalizeDeg(ad);
      deg += ad * turnRatio * turnSpeed;
      Math.normalizeDeg(deg);
    }
  }

  private void fireNormal() {
    if (padInput.button & PadState.Button.A) {
      turnRatio += (SLOW_TURN_RATIO - turnRatio) * TURN_CHANGE_RATIO;
      fireInterval = FIRE_INTERVAL;
      if (!aPressed) {
        fireCnt = 0;
        aPressed = true;
      }
    } else {
      turnRatio += (TURN_RATIO_BASE - turnRatio) * TURN_CHANGE_RATIO;
      aPressed = false;
      fireInterval *= 1.033f;
      if (fireInterval > FIRE_INTERVAL_MAX)
        fireInterval = 99999;
    }
    fireDeg = deg;
    if (reverseFire)
      fireDeg += PI;
    if (fireCnt <= 0) {
      SoundManager.playSe("shot.wav");
      Shot s = shots.getInstance();
      int foc = (fireSprCnt % 2) * 2 - 1;
      firePos.x = pos.x + cos(fireDeg + PI) * 0.2f * foc;
      firePos.y = pos.y - sin(fireDeg + PI) * 0.2f * foc;
      if (s)
        s.set(firePos, fireDeg);
      fireCnt = cast(int) fireInterval;
      float td;
      switch (foc) {
      case -1:
        td = fireSprDeg * (fireSprCnt / 2 % 4 + 1) * 0.2f;
        break;
      case 1:
        td = -fireSprDeg * (fireSprCnt / 2 % 4 + 1) * 0.2f;
        break;
      }
      fireSprCnt++;
      s = shots.getInstance();
      if (s)
        s.set(firePos, fireDeg + td);
      Smoke sm = smokes.getInstanceForced();
      float sd = fireDeg + td / 2;
      sm.set(firePos, sin(sd) * Shot.SPEED * 0.33f, cos(sd) * Shot.SPEED * 0.33f, 0,
             Smoke.SmokeType.SPARK, 10, 0.33f);
    }
    fireCnt--;
    if (padInput.button & PadState.Button.B) {
      if (!bPressed && fireLanceCnt <= 0 && !shots.existsLance()) {
        SoundManager.playSe("lance.wav");
        float fd = deg;
        if (reverseFire)
          fd += PI;
        Shot s = shots.getInstance();
        if (s)
          s.set(pos, fd, true);
        for (int i = 0; i < 4; i++) {
          Smoke sm = smokes.getInstanceForced();
          float sd = fd + rand.nextSignedFloat(1);
          sm.set(pos,
                 sin(sd) * Shot.LANCE_SPEED * i * 0.2f,
                 cos(sd) * Shot.LANCE_SPEED * i * 0.2f,
                 0, Smoke.SmokeType.SPARK, 15, 0.5f);
        }
        fireLanceCnt = FIRE_LANCE_INTERVAL;
      }
      bPressed = true;
    } else {
      bPressed = false;
    }
    fireLanceCnt--;
  }

  private void fireTwinStick() {
    if (fabs(stickInput.right.x) + fabs(stickInput.right.y) > 0.01f) {
      fireDeg = atan2(stickInput.right.x, stickInput.right.y);
      assert(fireDeg <>= 0);
      if (fireCnt <= 0) {
        SoundManager.playSe("shot.wav");
        int foc = (fireSprCnt % 2) * 2 - 1;
        float rsd = stickInput.right.vctSize;
        if (rsd > 1)
          rsd = 1;
        fireSprDeg = 1 - rsd + 0.05f;
        firePos.x = pos.x + cos(fireDeg + PI) * 0.2f * foc;
        firePos.y = pos.y - sin(fireDeg + PI) * 0.2f * foc;
        fireCnt = cast(int) fireInterval;
        float td;
        switch (foc) {
        case -1:
          td = fireSprDeg * (fireSprCnt / 2 % 4 + 1) * 0.2f;
          break;
        case 1:
          td = -fireSprDeg * (fireSprCnt / 2 % 4 + 1) * 0.2f;
          break;
        }
        fireSprCnt++;
        Shot s = shots.getInstance();
        if (s)
          s.set(firePos, fireDeg + td / 2, false, 2);
        s = shots.getInstance();
        if (s)
          s.set(firePos, fireDeg + td, false, 2);
        Smoke sm = smokes.getInstanceForced();
        float sd = fireDeg + td / 2;
        sm.set(firePos, sin(sd) * Shot.SPEED * 0.33f, cos(sd) * Shot.SPEED * 0.33f, 0,
               Smoke.SmokeType.SPARK, 10, 0.33f);
      }
    } else {
      fireDeg = 99999;
    }
    fireCnt--;
  }
  
  public bool checkBulletHit(Vector p, Vector pp) {
    if (cnt <= 0)
      return false;
    float bmvx, bmvy, inaa;
    bmvx = pp.x;
    bmvy = pp.y;
    bmvx -= p.x;
    bmvy -= p.y;
    inaa = bmvx * bmvx + bmvy * bmvy;
    if (inaa > 0.00001) {
      float sofsx, sofsy, inab, hd;
      sofsx = _pos.x;
      sofsy = _pos.y;
      sofsx -= p.x;
      sofsy -= p.y;
      inab = bmvx * sofsx + bmvy * sofsy;
      if (inab >= 0 && inab <= inaa) {
        hd = sofsx * sofsx + sofsy * sofsy - inab * inab / inaa;
        if (hd >= 0 && hd <= HIT_WIDTH) {
          destroyed();
          return true;
        }
      }
    }
    return false;
  }

  private void destroyed() {
    if (cnt <= 0)
      return;
    for (int i = 0; i < 128; i++) {
      Spark sp = sparks.getInstanceForced();
      sp.set(pos, rand.nextSignedFloat(1), rand.nextSignedFloat(1),
             0.5f + rand.nextFloat(0.5f), 0.5f + rand.nextFloat(0.5f), 0,
             40 + rand.nextInt(40));
    }
    if (shieldCnt > 0) {
      SoundManager.playSe("ship_shield_lost.wav");
      screen.setScreenShake(30, 0.02f);
      shieldCnt = 0;
      cnt = -INVINCIBLE_CNT / 2;
      return;
    }
    SoundManager.playSe("ship_destroyed.wav");
    for (int i = 0; i < 64; i++) {
      Smoke s = smokes.getInstanceForced();
      s.set(pos, rand.nextSignedFloat(0.2f), rand.nextSignedFloat(0.2f),
            rand.nextFloat(0.1f),
            Smoke.SmokeType.EXPLOSION, 50 + rand.nextInt(30), 1);
    }
    screen.setScreenShake(60, 0.05f);
    gameState.shipDestroyed();
    restart();
    cnt = -RESTART_CNT;
  }

  public bool hasCollision() {
    if (cnt < -INVINCIBLE_CNT)
      return false;
    else
      return true;
  }

  public void clearBullets() {
    gameState.clearBullets();
  }

  public void draw() {
    if (cnt < -INVINCIBLE_CNT)
      return;
    if (fireDeg < 99999) {
      Screen.setColor(0.5f, 0.9f, 0.7f, 0.3f);
      glBegin(GL_LINE_STRIP);
      glVertex2f(pos.x, pos.y);
      Screen.setColor(0.5f, 0.9f, 0.7f, 0.7f);
      glVertex2f(pos.x + sin(fireDeg) * 20, pos.y + cos(fireDeg) * 20);
      glEnd();
    }
    if (cnt < 0 && (-cnt % 32) < 16)
      return;
    glPushMatrix();
    Screen.glTranslate(pos);
    glRotatef(-deg * 180 / PI, 0, 0, 1);
    _shape.draw();
    bridgeShape.draw();
    if (shieldCnt > 0) {
      float ss = 0.66f;
      if (shieldCnt < 120)
        ss *= cast(float) shieldCnt / 120;
      glScalef(ss, ss, ss);
      glRotatef(shieldCnt * 5, 0, 0, 1);
      shieldShape.draw();
    }
    glPopMatrix();
  }

  public void drawShape() {
    _shape.draw();
    bridgeShape.draw();
  }

  public Vector pos() {
    return _pos;
  }

  public Vector vel() {
    return _vel;
  }

  public float scrollSpeedBase() {
    return _scrollSpeedBase;
  }

  public void setReplayMode(float turnSpeed, bool reverseFire) {
    _replayMode = true;
    this.turnSpeed = turnSpeed;
    this.reverseFire = reverseFire;
  }

  public void unsetReplayMode() {
    _replayMode = false;
    turnSpeed = GameManager.shipTurnSpeed;
    reverseFire = GameManager.shipReverseFire;
  }

  public bool replayMode() {
    return _replayMode;
  }
}
